local Promise = require(script.Promise.Value)
local Binder = require(script.Binder.Value)

local Players = game:GetService("Players")
local getHumanoidDescriptionFromUserIdPromise = Promise.promisify(function (...)
	return Players:GetHumanoidDescriptionFromUserId(...)
end)

local UserService = game:GetService("UserService")
local getUserInfosByUserIdsPromise = Promise.promisify(function (...)
	return UserService:GetUserInfosByUserIdsAsync(...)
end)

local PlayerSpoof = {}
PlayerSpoof.__index = PlayerSpoof

PlayerSpoof.TAG = "PlayerSpoof"
PlayerSpoof.ATTR_PREFIX = "PlayerSpoof"
PlayerSpoof.ATTR_DONT_AUTO_LOAD = PlayerSpoof.ATTR_PREFIX .. "DontAutoLoad" 
PlayerSpoof.ATTR_USER_ID = PlayerSpoof.ATTR_PREFIX .. "UserId"

PlayerSpoof.Binder = Binder.new(PlayerSpoof.TAG, PlayerSpoof)

function PlayerSpoof.new(figure)
	local self = setmetatable({
		human = assert(figure:FindFirstChildWhichIsA("Humanoid"));
		config = figure;
		_defaultHumanoidDescription = nil;
		_defaultDisplayName = nil;
	}, PlayerSpoof)
	self:autoLoad()
	self._userIdChangedConn = self.config:GetAttributeChangedSignal(PlayerSpoof.ATTR_USER_ID):Connect(function ()
		self:autoLoad()
	end)
	return self
end

function PlayerSpoof:destroy()
	self:revertChanges()
	self._userIdChangedConn:Disconnect()
	self._userIdChangedConn = nil
end

function PlayerSpoof:shouldAutoLoad()
	return not self.config:GetAttribute(PlayerSpoof.ATTR_DONT_AUTO_LOAD)
end

function PlayerSpoof:getUserId()
	return assert(
		self.config:GetAttribute(PlayerSpoof.ATTR_USER_ID),
		("%s missing attribute %s"):format(
			self.config:GetFullName(),
			PlayerSpoof.ATTR_USER_ID
		)
	)
end

function PlayerSpoof:getBackupHumanoidDescription()
	return self.config:FindFirstChild("BackupHumanoidDescription")
end

function PlayerSpoof:applyHumanoidDescription(humanoidDescription)
	if not self._defaultHumanoidDescription then
		self._defaultHumanoidDescription = self.human:GetAppliedDescription() 
	end
	self.human:ApplyDescription(humanoidDescription)
end

function PlayerSpoof:loadHumanoidDescription(userId)
	return getHumanoidDescriptionFromUserIdPromise(userId):andThen(function (humanoidDescription)
		self:applyHumanoidDescription(humanoidDescription)
	end):catch(function (err)
		local backupHumanoidDescription = self:getBackupHumanoidDescription()
		if backupHumanoidDescription then
			warn(("%s failed to load HumanoidDescription for UserId=%d, using backup %s"):format(
				self.figure:GetFullName(),
				userId,
				backupHumanoidDescription:GetFullName())
			)
			self:applyDescription(backupHumanoidDescription)
		else
			warn(("%s failed to load HumanoidDescription for UserId=%d, no backup"):format(
				self.figure:GetFullName(),
				userId)
			)
		end
	end)
end

function PlayerSpoof:setDisplayName(name)
	if not self._defaultDisplayName then
		self._defaultDisplayName = name
	end
	self.human.DisplayName = name or ""
end

function PlayerSpoof:loadDisplayName(userId)
	return getUserInfosByUserIdsPromise({userId}):andThen(function (infoArray)
		local found = false
		for i, t in pairs(infoArray) do
			if t["Id"] == userId then
				self:setDisplayName(t["DisplayName"] or t["UserName"])
				found = true
				break
			end
		end
		assert(found)
	end):catch(function (err)
		self:setDisplayName()
		warn(("%s failed to load DisplayName for UserId=%d"):format(userId))
	end)
end

function PlayerSpoof:load()
	if self._loadPromise then return self._loadPromise end
	local userId = self:getUserId()
	self:revertChanges()
	self._loadPromise = Promise.all{
		self:loadHumanoidDescription(userId);
		self:loadDisplayName(userId);
	}:finally(function ()
		self._loadPromise = nil
	end)
	return self._loadPromise
end

function PlayerSpoof:autoLoad()
	if not self:shouldAutoLoad() then return end
	self:load()
end

function PlayerSpoof:revertChanges()
	if self._defaultHumanoidDescription then
		self:applyHumanoidDescription(self._defaultHumanoidDescription)
		self._defaultHumanoidDescription = nil
	end
	if self._defaultDisplayName then
		self:setDisplayName(self._defaultDisplayName)
		self._defaultDisplayName = nil
	end
end

return PlayerSpoof
